from homeconnect import HomeConnect
import webbrowser
import time

def print_status(app):
    print(app.name, app.status)



if __name__ == '__main__':
    client_id = input("Please enter the client ID: ")
    client_secret = input("Please enter the client secret: ")
    redirect_uri = input("Please enter the redirect URI: ")

    hc = HomeConnect(client_id, client_secret, redirect_uri, simulate=True)

    webbrowser.open(hc.get_authurl())

    auth_result = input("Please enter the URL redirected to: ")

    hc.get_token(auth_result)

    appliances = hc.get_appliances()

    for app in appliances:
        app.get_status()
        print_status(app)
        app.listen_events(callback=print_status)



"""Taken from https://github.com/btubbs/sseclient"""

import codecs
import logging
import re
import time
import warnings

import requests
from oauthlib.oauth2 import TokenExpiredError
from requests.exceptions import HTTPError

# Technically, we should support streams that mix line endings.  This regex,
# however, assumes that a system will provide consistent line endings.
end_of_field = re.compile(r"\r\n\r\n|\r\r|\n\n")

LOGGER = logging.getLogger("homeconnect.sseclient")


class SSEClient(object):
    def __init__(
        self, url, last_id=None, retry=3000, session=None, chunk_size=1024, **kwargs
    ):
        self.url = url
        self.last_id = last_id
        self.retry = retry
        self.chunk_size = chunk_size

        # Optional support for passing in a requests.Session()
        self.session = session

        # Any extra kwargs will be fed into the requests.get call later.
        self.requests_kwargs = kwargs

        # The SSE spec requires making requests with Cache-Control: nocache
        if "headers" not in self.requests_kwargs:
            self.requests_kwargs["headers"] = {}
        self.requests_kwargs["headers"]["Cache-Control"] = "no-cache"

        # The 'Accept' header is not required, but explicit > implicit
        self.requests_kwargs["headers"]["Accept"] = "text/event-stream"

        # Keep data here as it streams in
        self.buf = ""

        self._connect()

    def _connect(self):
        LOGGER.info("Connecting ...")
        if self.last_id:
            self.requests_kwargs["headers"]["Last-Event-ID"] = self.last_id

        # Use session if set.  Otherwise fall back to requests module.
        requester = self.session or requests
        self.resp = requester.get(self.url, stream=True, **self.requests_kwargs)
        self.resp_iterator = self.resp.iter_content(chunk_size=self.chunk_size)

        self.resp.encoding = "UTF-8"

        # TODO: Ensure we're handling redirects.  Might also stick the 'origin'
        # attribute on Events like the Javascript spec requires.
        try:
            self.resp.raise_for_status()
        except HTTPError:
            LOGGER.error("Failed connecting.")
            # Wait 10 times longer if connection failed due to rate limits
            time.sleep(10 * self.retry / 1000.0)
            self._connect()

    def _event_complete(self):
        return re.search(end_of_field, self.buf) is not None

    def __iter__(self):
        return self

    def __next__(self):
        while not self._event_complete():
            try:
                decoder = codecs.getincrementaldecoder(self.resp.encoding)(
                    errors="replace"
                )
                next_chunk = next(self.resp_iterator)
                if not next_chunk:
                    LOGGER.error("EOFError")
                    raise EOFError()
                self.buf += decoder.decode(next_chunk)

            # except (StopIteration, requests.RequestException, EOFError, http.client.IncompleteRead, ValueError) as e:
            except Exception as e:
                LOGGER.warning("Exception while reading event: ", exc_info=True)
                time.sleep(self.retry / 1000.0)
                self._connect()

                # The SSE spec only supports resuming from a whole message, so
                # if we have half a message we should throw it out.
                head, sep, tail = self.buf.rpartition("\n")
                self.buf = head + sep
                continue

        # Split the complete event (up to the end_of_field) into event_string,
        # and retain anything after the current complete event in self.buf
        # for next time.
        (event_string, self.buf) = re.split(end_of_field, self.buf, maxsplit=1)
        msg = Event.parse(event_string)

        # If the server requests a specific retry delay, we need to honor it.
        if msg.retry:
            self.retry = msg.retry

        # last_id should only be set if included in the message.  It's not
        # forgotten if a message omits it.
        if msg.id:
            self.last_id = msg.id

        return msg


class Event(object):

    sse_line_pattern = re.compile("(?P<name>[^:]*):?( ?(?P<value>.*))?")

    def __init__(self, data="", event="message", id=None, retry=None):
        self.data = data
        self.event = event
        self.id = id
        self.retry = retry

    def dump(self):
        lines = []
        if self.id:
            lines.append("id: %s" % self.id)

        # Only include an event line if it's not the default already.
        if self.event != "message":
            lines.append("event: %s" % self.event)

        if self.retry:
            lines.append("retry: %s" % self.retry)

        lines.extend("data: %s" % d for d in self.data.split("\n"))
        return "\n".join(lines) + "\n\n"

    @classmethod
    def parse(cls, raw):
        """
        Given a possibly-multiline string representing an SSE message, parse it
        and return a Event object.
        """
        msg = cls()
        for line in raw.splitlines():
            m = cls.sse_line_pattern.match(line)
            if m is None:
                # Malformed line.  Discard but warn.
                warnings.warn('Invalid SSE line: "%s"' % line, SyntaxWarning)
                LOGGER.warn('Invalid SSE line: "%s"', line)
                continue

            name = m.group("name")
            if name == "":
                # line began with a ":", so is a comment.  Ignore
                continue
            value = m.group("value")

            if name == "data":
                # If we already have some data, then join to it with a newline.
                # Else this is it.
                if msg.data:
                    msg.data = "%s\n%s" % (msg.data, value)
                else:
                    msg.data = value
            elif name == "event":
                msg.event = value
            elif name == "id":
                msg.id = value
            elif name == "retry":
                msg.retry = int(value)

        return msg

    def __str__(self):
        return self.data



"""Tests for api classes."""

from requests.adapters import HTTPAdapter
import responses
from responses import registries

from homeconnect.api import ENDPOINT_APPLIANCES, TOTAL_RETRIES, URL_API, HomeConnectAPI

return_200_message = {
    "data": {
        "homeappliances": [
            {
                "name": "My Bosch Oven",
                "brand": "BOSCH",
                "vib": "HNG6764B6",
                "connected": True,
                "type": "Oven",
                "enumber": "HNG6764B6/09",
                "haId": "BOSCH-HNG6764B6-0000000011FF",
            }
        ]
    },
}


return_429_message = {
    "error": {
        "key": "429",
        "description": "The user has sent too many requests in a given amount of time.",
    }
}


def test_api() -> None:
    hc_api = HomeConnectAPI()
    assert hc_api.client_id is None


@responses.activate(registry=registries.OrderedRegistry)
def test_429_retry():
    """Test retries on 429 response status."""
    appliances_url = URL_API + ENDPOINT_APPLIANCES

    rsp_objs = [
        responses.get(
            appliances_url,
            json=return_429_message,
            status=429,
            headers={"Retry-After": "1"},
        )
        for _ in range(TOTAL_RETRIES)
    ]

    rsp_objs.append(responses.get(appliances_url, json=return_200_message, status=200))

    hc_api = HomeConnectAPI()

    adapter = HTTPAdapter(max_retries=hc_api.retry)

    hc_api._oauth.mount("https://", adapter)

    resp = hc_api.request("get", ENDPOINT_APPLIANCES)

    assert resp.status_code == 200
    for rsp in rsp_objs:
        assert rsp.call_count == 1


@responses.activate(registry=registries.OrderedRegistry)
def test_429_max_retry_exception():
    """Test max retries on 429 response status."""
    appliances_url = URL_API + ENDPOINT_APPLIANCES

    rsp_objs = [
        responses.get(
            appliances_url,
            json=return_429_message,
            status=429,
            headers={"Retry-After": "1"},
        )
        for _ in range(TOTAL_RETRIES)
    ]

    # Add one more than the number of retries.
    rsp_objs.append(
        responses.get(
            appliances_url,
            json=return_429_message,
            status=429,
            headers={"Retry-After": "1"},
        )
    )

    # Will not be called.
    rsp_objs.append(responses.get(appliances_url, json=return_200_message, status=200))

    hc_api = HomeConnectAPI()

    adapter = HTTPAdapter(max_retries=hc_api.retry)

    hc_api._oauth.mount("https://", adapter)

    resp = hc_api.request("get", ENDPOINT_APPLIANCES)

    assert rsp_objs[-1].call_count == 0
    assert resp is None



import json
import logging
import os
import time
from threading import Thread
from typing import Callable, Dict, Optional, Union

from oauthlib.oauth2 import TokenExpiredError
from requests import Response
from requests_oauthlib import OAuth2Session
from requests.exceptions import RetryError
from requests.adapters import HTTPAdapter, Retry

from .sseclient import SSEClient

URL_API = "https://api.home-connect.com"
ENDPOINT_AUTHORIZE = "/security/oauth/authorize"
ENDPOINT_TOKEN = "/security/oauth/token"
ENDPOINT_APPLIANCES = "/api/homeappliances"
TIMEOUT_S = 120
TOTAL_RETRIES = 1


LOGGER = logging.getLogger("homeconnect")


class HomeConnectError(Exception):
    pass


class HomeConnectAPI:
    def __init__(
        self,
        token: Optional[Dict[str, str]] = None,
        client_id: str = None,
        client_secret: str = None,
        redirect_uri: str = None,
        api_url: Optional[str] = None,
        token_updater: Optional[Callable[[str], None]] = None,
    ):
        self.host = api_url or URL_API
        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri
        self.token_updater = token_updater

        self._appliances = {}
        self.listening_events = False

        extra = {"client_id": self.client_id, "client_secret": self.client_secret}

        self._oauth = OAuth2Session(
            client_id=client_id,
            redirect_uri=redirect_uri,
            auto_refresh_kwargs=extra,
            token=token,
            token_updater=token_updater,
        )
        self.retry = Retry(TOTAL_RETRIES, status_forcelist=[429])
        self._oauth.mount("https://", HTTPAdapter(max_retries=self.retry))
        self._oauth.mount("http://", HTTPAdapter(max_retries=self.retry))

    def refresh_tokens(self) -> Dict[str, Union[str, int]]:
        """Refresh and return new tokens."""
        LOGGER.info("Refreshing tokens ...")
        token = self._oauth.refresh_token(f"{self.host}{ENDPOINT_TOKEN}")

        if self.token_updater is not None:
            self.token_updater(token)

        return token

    def request(self, method: str, path: str, **kwargs) -> Response:
        """Make a request.

        We don't use the built-in token refresh mechanism of OAuth2 session because
        we want to allow overriding the token refresh logic.
        """
        url = f"{self.host}/{path.lstrip('/')}"
        try:
            return getattr(self._oauth, method)(url, **kwargs)
        except TokenExpiredError:
            LOGGER.warning("Token expired.")
            self._oauth.token = self.refresh_tokens()

            return getattr(self._oauth, method)(url, **kwargs)
        except RetryError as e:
            LOGGER.warning("Retry failed: %s", e)
            return e.response

    def get(self, endpoint):
        """Get data as dictionary from an endpoint."""
        res = self.request("get", endpoint)
        if not res.content:
            return {}
        try:
            res = res.json()
        except:
            raise ValueError("Cannot parse {} as JSON".format(res))
        if "error" in res:
            raise HomeConnectError(res["error"])
        elif "data" not in res:
            raise HomeConnectError("Unexpected error")
        return res["data"]

    def put(self, endpoint, data):
        """Send (PUT) data to an endpoint."""
        res = self.request(
            "put",
            endpoint,
            data=json.dumps(data),
            headers={
                "Content-Type": "application/vnd.bsh.sdk.v1+json",
                "accept": "application/vnd.bsh.sdk.v1+json",
            },
        )
        if not res.content:
            return {}
        try:
            res = res.json()
        except:
            raise ValueError("Cannot parse {} as JSON".format(res))
        if "error" in res:
            raise HomeConnectError(res["error"])
        return res

    def delete(self, endpoint):
        """Delete an endpoint."""
        res = self.request("delete", endpoint)
        if not res.content:
            return {}
        try:
            res = res.json()
        except:
            raise ValueError("Cannot parse {} as JSON".format(res))
        if "error" in res:
            raise HomeConnectError(res["error"])
        return res

    def get_appliances(self):
        """Return a list of `HomeConnectAppliance` instances for all
        appliances."""

        appliances = {}

        data = self.get(ENDPOINT_APPLIANCES)
        for home_appliance in data["homeappliances"]:
            haId = home_appliance["haId"]

            if haId in self._appliances:
                appliances[haId] = self._appliances[haId]
                appliances[haId].connected = home_appliance["connected"]
                continue

            appliances[haId] = HomeConnectAppliance(self, **home_appliance)

        self._appliances = appliances
        return list(self._appliances.values())

    def get_authurl(self):
        """Get the URL needed for the authorization code grant flow."""
        authorization_url, _ = self._oauth.authorization_url(
            f"{self.host}{ENDPOINT_AUTHORIZE}"
        )
        return authorization_url

    def listen_events(self):
        """Spawn a thread with an event listener that updates the status."""
        self.listening_events = True
        uri = f"{self.host}/api/homeappliances/events"
        sse = SSEClient(uri, session=self._oauth, retry=1000, timeout=TIMEOUT_S)
        Thread(target=self._listen, args=[sse]).start()

    def _listen(self, sse):
        """Worker function for listener."""
        LOGGER.info("Listening to event stream for all devices")
        try:
            for event in sse:
                try:
                    for appliance in self._appliances.values():
                        if appliance.haId == event.id:
                            self.handle_event(event, appliance)
                            break
                except ValueError:
                    pass
        except TokenExpiredError:
            LOGGER.info("Token expired in event stream.")
            self._oauth.token = self.refresh_tokens()
            uri = f"{self.host}/api/homeappliances/events"
            sse = SSEClient(uri, session=self._oauth, retry=1000, timeout=TIMEOUT_S)
            self._listen(sse)

    def handle_event(self, event, appliance):
        """Handle a new event.

        Updates the status with the event data and executes any callback
        function."""
        event_data = json.loads(event.data)
        items = event_data.get("items")
        if items is not None:
            data_dict = self.json2dict(items)
        else:
            data_dict = {event_data.pop("key"): event_data}

        if event.event in ("NOTIFY", "STATUS", "EVENT"):
            appliance.status.update(data_dict)

        elif event.event == "CONNECTED":
            appliance.connected = True
        elif event.event == "DISCONNECTED":
            appliance.connected = False

        if appliance.event_callback is not None:
            try:
                appliance.event_callback(appliance, event.event, data_dict)
            except TypeError:
                appliance.event_callback(appliance)

    @staticmethod
    def json2dict(lst):
        """Turn a list of dictionaries where one key is called 'key'
        into a dictionary with the value of 'key' as key."""
        return {d.pop("key"): d for d in lst}


class HomeConnect(HomeConnectAPI):
    """Connection to the HomeConnect OAuth API."""

    def __init__(
        self,
        client_id,
        client_secret="",
        redirect_uri="",
        api_url: Optional[str] = None,
        token_cache=None,
    ):
        """Initialize the connection."""
        self.token_cache = token_cache or "homeconnect_oauth_token.json"
        super().__init__(
            None, client_id, client_secret, redirect_uri, api_url, self.token_dump
        )

    def token_dump(self, token):
        """Dump the token to a JSON file."""
        with open(self.token_cache, "w") as f:
            json.dump(token, f)

    def token_load(self):
        """Load the token from the cache if exists it and is not expired,
        otherwise return None."""
        if not os.path.exists(self.token_cache):
            return None
        with open(self.token_cache, "r") as f:
            token = json.load(f)
        now = int(time.time())
        token["expires_in"] = token.get("expires_at", now - 1) - now
        self._oauth = OAuth2Session(
            client_id=self.client_id,
            redirect_uri=self.redirect_uri,
            auto_refresh_kwargs={"client_id": self.client_id, "client_secret": self.client_secret},
            token=token,
            token_updater=self.token_updater,
        )
        return token

    def token_expired(self, token):
        """Check if the token is expired."""
        now = int(time.time())
        return token["expires_at"] - now < 60

    def get_token(self, authorization_response):
        """Get the token given the redirect URL obtained from the
        authorization."""
        LOGGER.info("Fetching token ...")
        token = self._oauth.fetch_token(
            f"{self.host}{ENDPOINT_TOKEN}",
            authorization_response=authorization_response,
            client_secret=self.client_secret,
        )
        self.token_dump(token)


class HomeConnectAppliance:
    """Class representing a single appliance."""

    def __init__(
        self,
        hc,
        haId,
        vib=None,
        brand=None,
        type=None,
        name=None,
        enumber=None,
        connected=False,
    ):
        self.hc = hc
        self.haId = haId
        self.vib = vib or ""
        self.brand = brand or ""
        self.type = type or ""
        self.name = name or ""
        self.enumber = enumber or ""
        self.connected = connected
        self.status = {}

        self.event_callback = None

    def __repr__(self):
        return "HomeConnectAppliance(hc, haId='{}', vib='{}', brand='{}', type='{}', name='{}', enumber='{}', connected={})".format(
            self.haId,
            self.vib,
            self.brand,
            self.type,
            self.name,
            self.enumber,
            self.connected,
        )

    def listen_events(self, callback=None):
        """Register event callback method"""
        self.event_callback = callback

        if not self.hc.listening_events:
            self.hc.listen_events()

    @staticmethod
    def json2dict(lst):
        """Turn a list of dictionaries where one key is called 'key'
        into a dictionary with the value of 'key' as key."""
        return {d.pop("key"): d for d in lst}

    def get(self, endpoint):
        """Get data (as dictionary) from an endpoint."""
        return self.hc.get("{}/{}{}".format(ENDPOINT_APPLIANCES, self.haId, endpoint))

    def delete(self, endpoint):
        """Delete endpoint."""
        return self.hc.delete(
            "{}/{}{}".format(ENDPOINT_APPLIANCES, self.haId, endpoint)
        )

    def put(self, endpoint, data):
        """Send (PUT) data to an endpoint."""
        return self.hc.put(
            "{}/{}{}".format(ENDPOINT_APPLIANCES, self.haId, endpoint), data
        )

    def get_programs_active(self):
        """Get active programs."""
        return self.get("/programs/active")

    def get_programs_selected(self):
        """Get selected programs."""
        return self.get("/programs/selected")

    def get_programs_available(self):
        """Get available programs."""
        programs = self.get("/programs/available")
        if not programs or "programs" not in programs:
            return []
        return [p["key"] for p in programs["programs"]]

    def get_program_options(self, program_key):
        """Get program options."""
        options = self.get(f"/programs/available/{program_key}")
        if not options or "options" not in options:
            return []
        return [{p["key"]: p} for p in options["options"]]

    def start_program(self, program_key, options=None):
        """Start a program."""
        if options is not None:
            return self.put(
                "/programs/active", {"data": {"key": program_key, "options": options}}
            )
        return self.put("/programs/active", {"data": {"key": program_key}})

    def stop_program(self):
        """Stop a program."""
        return self.delete("/programs/active")

    def select_program(self, program, options=None):
        """Select a program."""
        if options is None:
            _options = {}
        else:
            _options = {"options": options}
        return self.put("/programs/selected", {"data": {"key": program, **_options}})

    def get_status(self):
        """Get the status (as dictionary) and update `self.status`."""
        status = self.get("/status")
        if not status or "status" not in status:
            return {}
        self.status = self.json2dict(status["status"])
        return self.status

    def get_settings(self):
        """Get the current settings."""
        settings = self.get("/settings")
        if not settings or "settings" not in settings:
            return {}
        self.status.update(self.json2dict(settings["settings"]))
        return self.status

    def set_setting(self, settingkey, value):
        """Change the current setting of `settingkey`."""
        return self.put(
            "/settings/{}".format(settingkey),
            {"data": {"key": settingkey, "value": value}},
        )

    def set_options_active_program(self, option_key, value, unit=None):
        """Change the option `option_key` of the currently active program."""
        if unit is None:
            _unit = {}
        else:
            _unit = {"unit": unit}
        return self.put(
            f"/programs/active/options/{option_key}",
            {"data": {"key": option_key, "value": value, **_unit}},
        )

    def set_options_selected_program(self, option_key, value, unit=None):
        """Change the option `option_key` of the currently selected program."""
        if unit is None:
            _unit = {}
        else:
            _unit = {"unit": unit}
        return self.put(
            f"/programs/selected/options/{option_key}",
            {"data": {"key": option_key, "value": value, **_unit}},
        )

    def execute_command(self, command):
        """Execute a command."""
        return self.put(
            f"/commands/{command}",
            {"data": {"key": command, "value": True}},
        )



from setuptools import find_packages, setup

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="homeconnect",
    version="0.8.0",
    author="David M. Straub",
    author_email="straub@protonmail.com",
    url="https://github.com/DavidMStraub/homeconnect",
    description="Python client for the BSH Home Connect REST API",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(),
    install_requires=["requests", "requests_oauthlib"],
    extras_require={
        "testing": [
            "nose",
        ],
    },
)
